--[[{{{

{
	type = "blueprint",
	version = "0.13.0",
	data = {
		label = "This is a blueprint",
		entities = {},
		icons = {},
		tiles = {}
	}
}



{
	type = "blueprint-book",
	version = "0.13.0",
	data = {
		label = "This is a blueprint book",
		active = {
			label = "This is a blueprint",
			entities = {},
			icons = {},
			tiles = {}
		},
		main = {
			[1] = {
				label = "This is a blueprint",
				entities = {},
				icons = {},
				tiles = {}
			},
			[2] = ...
		}
	}
}

}}}]]--

require "core"

local M = {}

-- =========================================================================
-- {[> Attributes
-- ==============  {{{
local base64			= require "libs.base64"
local blueprintstring	= require "libs.blueprintstring"
local deflate			= require "libs.zlib-deflate"
local inflate			= require "libs.deflatelua"
local serpent			= require "libs.serpent_0-28"

local gameVersion
--}}}



-- =========================================================================
-- {[> Local Methods
-- =================
function deleteUnneededFields(t) --{{{

	if (not t) then
		return
	end

	for _, entity in ipairs(t) do
		entity.entity_number = nil
	end
	
end --}}}

function getDataStructure(t) --{{{

	if t and (type(t) == "table") then
	
		if t[1] and t[1].type and (type(t[1].type) == "string") then
			return "collection"
		end
	
		if t.type and (type(t.type) == "string") then
			if (t.type == "blueprint") then
				return "blueprint"
			elseif (t.type == "blueprint-book") then
				return "blueprint-book"
			end
		end
		
		if t.entities and (type(t.entities) == "table") then
			return "raw"
		end
		
		if t.tiles and (type(t.tiles) == "table") then
			return "raw"
		end
		
	end
	
	return nil

end --}}}

function recreateEntities(t) --{{{

	if (not t) then
		return nil
	end

	local entities = {}
	local c = 1

	for _, entity in ipairs(t) do
		entity.entity_number = c
		entities[c] = entity
		c = c + 1
	end
	
	return entities

end --}}}



-- =========================================================================
-- {[> Module Methods
-- ==================
M.fromString = function(origString) --{{{

	local status, result
	
	-- decode base64
	local string = base64.dec(origString:trim())
	
	-- uncompress string
	local decompressedData = {}
	status, result = pcall(inflate.gunzip, {
		input = string,
		output = function(byte)
			decompressedData[#(decompressedData) + 1] = string.char(byte)
		end
	})
	
	if (not status) then
		return nil
	end
	
	string = table.concat(decompressedData)
	
	-- create table from string
	status, result = serpent.load(string)
	
	if (not status) then
		return nil
	end
	
	-- determine data structure
	local structure = getDataStructure(result)
	
	if (not structure) then
		return nil
	end
	
	-- recreate entity indexes
	if (structure == "blueprint") then
		result.data.entities = recreateEntities(result.data.entities)
	elseif (structure == "blueprint-book") then
		if result.data.active then
			result.data.active.entities = recreateEntities(result.data.active.entities)
		end
		if result.data.main then
			for i, t in ipairs(result.data.main) do
				result.data.main[i].entities = recreateEntities(t.entities)
			end
		end
	elseif (structure == "collection") then
		for i, entry in ipairs(result) do
			if (entry.type == "blueprint") then
				result[i].data.entities = recreateEntities(entry.data.entities)
			elseif (entry.type == "blueprint-book") then
				if result[i].data.active then
					result[i].data.active.entities = recreateEntities(entry.data.active.entities)
				end
				if entry.data.main then
					for j, t in ipairs(entry.data.main) do
						result[i].data.main[j].entities = recreateEntities(t.entities)
					end
				end
			else
				return nil
			end
		end
	elseif (structure == "raw") then
		
		result = blueprintstring.fromString(string)
		
		local newResult = {
			type = "blueprint",
			version = "0.13.0",
			data = {
				entities = result.entities,
				icons = result.icons,
				tiles = result.tiles
			}
		}
		
		result = newResult
		structure = "blueprint"
	
	end
	
	
	
	-- return table and type
	return result, structure

end --}}}

M.getGameVersion = function() --{{{

	if (gameVersion == nil) then
		for k, v in ipairs(game.active_mods) do
			if (k == "base") then
				gameVersion = v
				break
			end
		end
	end
	
	return gameVersion
	
end --}}}

M.toString = function(data) --{{{

	data = table.deepcopy(data)

	-- determine data structure
	local structure = getDataStructure(data)
	if (not structure) then
		return nil
	end
	
	
	
	-- remove unneeded fields to shorten the resulting string
	if (structure == "blueprint") then
		deleteUnneededFields(data.data.entities)
	elseif (structure == "blueprint-book") then
		if data.data.active then
			deleteUnneededFields(data.data.active.entities)
		end
		if data.data.main then
			for _, t in ipairs(data.data.main) do
				deleteUnneededFields(t.entities)
			end
		end
	elseif (structure == "collection") then
		for _, entry in ipairs(data) do
			if (entry.type == "blueprint") then
				deleteUnneededFields(entry.data.entities)
			elseif (entry.type == "blueprint-book") then
				if entry.data.active then
					deleteUnneededFields(entry.data.active.entities)
				end
				if entry.data.main then
					for _, t in ipairs(entry.data.main) do
						deleteUnneededFields(t.entities)
					end
				end
			else
				return nil
			end
		end
	else
		return nil
	end
	
	
	
	-- create, compress and return string
	local s = serpent.dump(data)
	s = deflate.gzip(s)
	s = base64.enc(s)
	s = s:gsub(("%S"):rep(128), "%1\n")
	
	return s

end --}}}



-- =========================================================================
-- {[> End of Module
-- =================
return M